(*
 * Simple IR evaluator.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ast_util
open Fc_ir_util
open Fc_ir_type
open Fc_ir_env
open Fc_ir_exn
open Fc_ir_exn_type

(************************************************************************
 * TYPES
 ************************************************************************)

(*
 * Values.
 *)
type ir_value =
   ValNone
 | ValChar of char ref
 | ValInt of int ref
 | ValFloat of float ref
 | ValPointer of (ir_value array * int) option ref
 | ValStruct of fields
 | ValFun of (int * (venv -> exp exn_loc -> ir_value list -> ir_value)) option ref

and fields = (var * ir_value) list

(*
 * Variable environment.
 *)
and venv = ir_value SymbolTable.t
type evalue = ir_value

(*
 * Type environment.
 *)
let raise_exn code v =
   IRException (code, v)

let tenv_lookup      = tenv_lookup raise_exn
let tenv_expand      = tenv_expand raise_exn

(*
 * Return a value from the function.
 *)
exception ReturnExn of venv * ir_value

(*
 * Evaluation exception.
 *)
exception EvalException of exp exn_loc * venv * string

(*
 * Use the debugger.
 *)
let debug_eval = ref false

(*
 * Program values.
 *)
let main_sym1 = Symbol.add "main"
let main_sym2 = Symbol.add "%main"
let exit_sym = Symbol.add "exit"

(************************************************************************
 * BUILTIN FUNCTIONS
 ************************************************************************)

(*
 * Make a normal string from the value.
 *)
let make_string_aux a i =
   let len = Array.length a in
      if i < 0 || i > len then
         raise (Failure "make_string_aux");
      let s = String.create (len - i) in
      let rec loop j i =
         if i <> len then
            match a.(i) with
               ValChar { contents = c } ->
                  s.[j] <- c;
                  loop (succ j) (succ i)
             | _ ->
                  raise (Failure "make_string_aux")
      in
         loop 0 i;
         s

let make_string venv pos a i =
   try make_string_aux a i with
      Failure _ ->
         raise (EvalException (pos, venv, "make_string: not a string"))

(*
 * Convert a string to a number.
 *)
let atoi venv pos = function
   [ValPointer { contents = Some (a, i) }] ->
      ValInt (ref (int_of_string (make_string venv pos a i)))
 | _ ->
      raise (EvalException (pos, venv, "atoi: not an int*"))

(*
 * Exit the program.
 *)
let do_exit venv pos = function
   [v] ->
      raise (ReturnExn (venv, v))
 | _ ->
      raise (EvalException (pos, venv, "exit: not an int"))

(*
 * Print a char.
 *)
let print_char venv pos = function
   [ValChar { contents = c }] ->
      print_char c;
      ValInt (ref 0)
 | _ ->
      raise (EvalException (pos, venv, "print_char: not a char"))

(*
 * Print an int.
 *)
let print_int venv pos = function
   [ValInt { contents = i }] ->
      print_int i;
      ValInt (ref 0)
 | _ ->
      raise (EvalException (pos, venv, "print_int: not an int"))

(*
 * Print a float.
 *)
let print_float venv pos = function
   [ValFloat { contents = x }] ->
      print_float x;
      ValInt (ref 0)
 | _ ->
      raise (EvalException (pos, venv, "print_float: not a float"))

(*
 * Print a string.
 *)
let print_string venv pos = function
   [ValPointer { contents = Some (a, i) }] ->
      let s = make_string venv pos a i in
         print_string s;
         ValInt (ref 0)
 | _ ->
      raise (EvalException (pos, venv, "print_string: not a string"))

(************************************************************************
 * PRINTER
 ************************************************************************)

(*
 * Print a value.
 *)
let rec print_val = function
   ValNone ->
      Format.print_string "$void"
 | ValChar { contents = c } ->
      Format.print_string ("'" ^ Char.escaped c ^ "'")
 | ValInt { contents = i } ->
      Format.print_int i
 | ValFloat { contents = x } ->
      Format.print_float x
 | ValPointer { contents = None } ->
      Format.print_string "nil"
 | ValPointer { contents = Some (a, i) } ->
      (try
          let s = make_string_aux a i in
             Format.print_string ("\"" ^ String.escaped s ^ "\"")
       with
          Failure _ ->
             Format.open_hvbox tabstop;
             Format.print_string "<pointer = &{ ";
             Array.iter (fun v ->
                   print_val v;
                   Format.print_string ";";
                   Format.print_space ()) a;
             Format.print_string "}[";
             Format.print_int i;
             Format.print_string "]>";
             Format.close_box ())
 | ValFun { contents = Some _ } ->
      Format.print_string "<fun>"
 | ValFun { contents = None } ->
      Format.print_string "<fun=nil>"
 | ValStruct fields ->
      Format.open_hvbox 0;
      Format.open_hvbox 2;
      Format.print_string "{";
      List.iter (fun (v1, v2) ->
            Format.print_space ();
            Format.open_hvbox tabstop;
            print_symbol v1;
            Format.print_string " =";
            Format.print_space ();
            print_val v2;
            Format.print_string ";";
            Format.close_box ()) fields;
      Format.close_box ();
      Format.print_space ();
      Format.print_string "}";
      Format.close_box ()

(*
 * Print the variable environment.
 *)
let print_venv venv =
   SymbolTable.iter (fun v a ->
         Format.open_hvbox tabstop;
         print_symbol v;
         Format.print_string " =";
         Format.print_space ();
         print_val a;
         Format.close_box ();
         Format.print_space ()) venv

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * Variable environment.
 *)
let builtin_vals =
   ["print_char",   (1, print_char);
    "print_int",    (1, print_int);
    "print_float",  (1, print_float);
    "print_string", (1, print_string);
    "atoi",         (1, atoi);
    "exit",         (1, do_exit)]

let add_normal venv vals =
   List.fold_left (fun venv (s, v) ->
         let v = ValFun (ref (Some v)) in
         let f1 = Symbol.add s in
         let f2 = Symbol.add ("%" ^ s) in
         let venv = SymbolTable.add venv f1 v in
            SymbolTable.add venv f2 v) venv vals

let empty_venv = add_normal (SymbolTable.create ()) builtin_vals

let venv_add venv v1 v2 =
   SymbolTable.add venv v1 v2

let venv_mem venv v =
   SymbolTable.mem venv v

let venv_lookup venv pos v =
   try SymbolTable.find venv v with
      Not_found ->
         raise (IRException (pos, UnboundVar v))

(*
 * Add a list of arguments.
 *)
let venv_add_args venv pos vars args =
   let len1 = List.length vars in
   let len2 = List.length args in
   let _ =
      if len1 <> len2 then
         raise (IRException (pos, ArityMismatch (len1, len2)))
   in
   let rec loop venv vars args =
      match vars, args with
         v :: vars, a :: args ->
            loop (venv_add venv v a) vars args
       | [], [] ->
            venv
       | _ ->
            raise (Invalid_argument "Fc_ir_eval.venv_add_args")
   in
      loop venv vars args

(************************************************************************
 * COERCIONS
 ************************************************************************)

(*
 * Test the value.
 * Argument should be an int.
 *)
let coerce_bool venv pos = function
   ValInt { contents = i } -> i <> 0
 | _ -> raise (EvalException (pos, venv, "coerce_bool: not an int"))

(*
 * Get the pointer.
 *)
let coerce_pointer venv pos = function
   ValPointer { contents = Some (a, i) } -> a, i
 | _ -> raise (EvalException (pos, venv, "coerce_pointer: not a pointer"))

(*
 * Get an int.
 *)
let coerce_int venv pos = function
   ValInt { contents = i } -> i
 | _ -> raise (EvalException (pos, venv, "coerce_int: not an int"))

(*
 * Structure fields.
 *)
let coerce_struct venv pos = function
   ValStruct s ->
      s
 | ValPointer { contents = Some (a, i) } ->
      if i < 0 || i > Array.length a then
         raise (EvalException (pos, venv, "array index out of bounds"));
      (match a.(i) with
          ValStruct s -> s
        | _ -> raise (EvalException (pos, venv, "coerce_struct: not a struct")))
 | _ ->
      raise (EvalException (pos, venv, "coerce_struct: not a struct"))

(************************************************************************
 * COPYING
 ************************************************************************)

(*
 * Default value.
 * Return random values to help catch bogus code.
 *)
let rec default_value tenv pos = function
   TyAny -> ValNone
 | TyChar -> ValChar (ref (Char.chr (Random.int 256)))
 | TyInt -> ValInt (ref (Random.bits ()))
 | TyFloat -> ValFloat (ref (Random.float 1e30))
 | TyArray (ty, Some i) ->
      let a = Array.init i (fun _ -> default_value tenv pos ty) in
         ValPointer (ref (Some (a, 0)))
 | TyArray (ty, None) ->
      ValPointer (ref None)
 | TyStruct fields ->
      ValStruct (List.map (fun (v, ty) -> v, default_value tenv pos ty) fields)
 | TyFun _ -> ValFun (ref None)
 | TyId id -> default_value tenv pos (tenv_lookup tenv pos id)

(*
 * Copy a value.
 *)
let rec copy_value = function
   ValNone -> ValNone
 | ValChar { contents = c } -> ValChar (ref c)
 | ValInt { contents = i } -> ValInt (ref i)
 | ValFloat { contents = x } -> ValFloat (ref x)
 | ValPointer { contents = Some (a, i) } ->
      ValPointer (ref (Some (Array.map copy_value a, i)))
 | ValPointer { contents = None } ->
      ValPointer (ref None)
 | ValStruct fields ->
      ValStruct (List.map (fun (v, v') -> v, copy_value v') fields)
 | ValFun { contents = f } -> ValFun (ref f)

(*
 * Copy a an atom.
 *)
let rec copy_atom = function
   ValNone -> ValNone
 | ValChar { contents = c } -> ValChar (ref c)
 | ValInt { contents = i } -> ValInt (ref i)
 | ValFloat { contents = x } -> ValFloat (ref x)
 | ValPointer { contents = Some (a, i) } ->
      ValPointer (ref (Some (a, i)))
 | ValPointer { contents = None } ->
      ValPointer (ref None)
 | ValStruct fields ->
      ValStruct (List.map (fun (v, v') -> v, copy_atom v') fields)
 | ValFun { contents = f } -> ValFun (ref f)

(*
 * Set a value.
 *)
let rec set_value venv pos v1 v2 =
   match v1, v2 with
      ValChar p, ValChar { contents = c } ->
         p := c
    | ValInt p, ValInt { contents = i } ->
         p := i
    | ValFloat p, ValFloat { contents = x } ->
         p := x
    | ValPointer p1, ValPointer { contents = p2 } ->
         p1 := p2
    | ValStruct s1, ValStruct s2 ->
         set_struct venv pos s1 s2
    | _ ->
         raise (EvalException (pos, venv, "set_value: type mismatch"))

and set_struct venv pos fields1 fields2 =
   match fields1, fields2 with
      (_, v1) :: fields1, (_, v2) :: fields2 ->
         set_value venv pos v1 v2
    | [], [] ->
         ()
    | _ ->
         raise (EvalException (pos, venv, "set_struct: fields don't have the same length"))

(*
 * Copy the values that the pointers point to.
 *)
let eval_deref venv pos a =
   let a, i = coerce_pointer venv pos a in
      if i < 0 || i >= Array.length a then
         raise (EvalException (pos, venv, "array index out of bounds"));
      a.(i)

let memcpy venv pos a1 a2 =
   let v1 = eval_deref venv pos a1 in
   let v2 = eval_deref venv pos a2 in
      set_value venv pos v1 v2

(************************************************************************
 * BINARY OPERATIONS
 ************************************************************************)

(*
 * Arith on pointers or numbers.
 *)
let eval_ptr_arith venv pos op_int op_float v1 v2 =
   match v1, v2 with
      ValInt { contents = i1 }, ValInt { contents = i2 } ->
         ValInt (ref (op_int i1 i2))
    | ValFloat { contents = x1 }, ValFloat { contents = x2 } ->
         ValFloat (ref (op_float x1 x2))
    | ValPointer { contents = Some (a, i) }, ValInt { contents = j } ->
         ValPointer (ref (Some (a, op_int i j)))
    | ValPointer { contents = Some (a1, i1) }, ValPointer { contents = Some (a2, i2) } ->
         if a1 != a2 then
            raise (EvalException (pos, venv, "eval_ptr_arith: pointers are not compatible"));
         ValInt (ref (op_int i1 i2))
    | ValPointer { contents = None }, _ ->
         raise (EvalException (pos, venv, "eval_ptr_arith: nil pointer"))
    | _ ->
         raise (EvalException (pos, venv, "eval_ptr_arith: type error"))

(*
 * Arith just on numbers.
 *)
let eval_arith venv pos op_int op_float v1 v2 =
   match v1, v2 with
      ValInt { contents = i1 }, ValInt { contents = i2 } ->
         ValInt (ref (op_int i1 i2))
    | ValFloat { contents = x1 }, ValFloat { contents = x2 } ->
         ValFloat (ref (op_float x1 x2))
    | _ ->
         raise (EvalException (pos, venv, "eval_arith: type error"))

(*
 * Arith just on ints.
 *)
let eval_int_arith venv pos op_int v1 v2 =
   match v1, v2 with
      ValInt { contents = i1 }, ValInt { contents = i2 } ->
         ValInt (ref (op_int i1 i2))
    | _ ->
         raise (EvalException (pos, venv, "eval_int_arith: type error"))

(*
 * Equality judgment.
 *)
let array_compare a1 a2 =
   let len1 = Array.length a1 in
   let len2 = Array.length a2 in
   let rec array_compare i =
      i = len1 || (a1.(i) == a2.(i) && array_compare (succ i))
   in
      len1 = len2 && array_compare 0

let eval_equal venv pos v1 v2 =
   match v1, v2 with
      ValChar { contents = c1 }, ValChar { contents = c2 } ->
         c1 = c2
    | ValInt { contents = i1 }, ValInt { contents = i2 } ->
         i1 = i2
    | ValFloat { contents = x1 }, ValFloat { contents = x2 } ->
         x1 = x2
    | ValPointer { contents = None }, ValPointer { contents = None } ->
         true
    | ValPointer { contents = Some (a1, i1) }, ValPointer { contents = Some (a2, i2) } ->
         i1 = i2 && array_compare a1 a2
    | ValPointer _, ValPointer _ ->
         false
    | _ ->
         raise (EvalException (pos, venv, "eval_equal: type_error"))

(*
 * Relations on pointers or numbers.
 *)
let eval_relation venv op_int op_float op_ptr pos v1 v2 =
   match v1, v2 with
      ValChar { contents = c1 }, ValChar { contents = c2 } ->
         op_int (Char.code c1) (Char.code c2)
    | ValInt { contents = i1 }, ValInt { contents = i2 } ->
         op_int i1 i2
    | ValFloat { contents = x1 }, ValFloat { contents = x2 } ->
         op_float x1 x2
    | ValPointer { contents = p1 }, ValPointer { contents = p2 } ->
         op_ptr p1 p2
    | _ ->
         raise (EvalException (pos, venv, "eval_relation: type error"))

(*
 * Pointer comparisons.
 *)
let ptr_compare op p1 p2 =
   match p1, p2 with
      None, None ->
         op 0 0
    | None, Some _
    | Some _, None ->
         false
    | Some (a1, i1), Some (a2, i2) ->
         op i1 i2 && array_compare a1 a2

(************************************************************************
 * DEBUGGER
 ************************************************************************)

(*
 * This is a really simple debugger.
 * It can print the code and examine variables.
 *)
let rec debugger venv code =
   output_string stdout "# ";
   flush stdout;
   match String_util.split " \t" (input_line stdin) with
      [] -> debugger venv code
    | ["l"] -> debugger_list venv code
    | ["p"; v] -> debugger_print venv code v
    | ["s"] -> ()
    | _ -> debugger_help venv code

(*
 * Print the code.
 *)
and debugger_list venv code =
   Format.set_formatter_out_channel stdout;
   print_expr code;
   Format.print_newline ()

(*
 * Print a variable.
 * Have to parse the suffix.
 *)
and debugger_print venv code s =
   let v = Symbol.debug_add s in
   let _ =
      try
         let v' = venv_lookup venv (atom_pos code) v in
            Format.set_formatter_out_channel stdout;
            print_val v';
            Format.print_newline ()
      with
         IRException _ ->
            Printf.printf "%s not bound\n" s
   in
      debugger venv code

(*
 * Debugger help message.
 *)
and debugger_help venv code =
   output_string stdout "Debugging commands:
\tl: list the current program
\tp <name>: print the value for a variable
\ts: continue execution\n"

(************************************************************************
 * EVALUATOR
 ************************************************************************)

(*
 * Make a string value.
 *)
let make_string_val s =
   let len = String.length s in
   let rec collect i =
      if i = len then
         []
      else
         ValChar (ref s.[i]) :: collect (succ i)
   in
   let a = Array.of_list (collect 0) in
      ValPointer (ref (Some (a, 0)))

(*
 * Evaluate an atom.
 *)
let eval_atom venv pos a =
   let pos = string_pos "eval_atom" pos in
      match a with
         AtomChar c -> ValChar (ref c)
       | AtomInt i -> ValInt (ref i)
       | AtomFloat x -> ValFloat (ref x)
       | AtomNil _ -> ValPointer (ref None)
       | AtomVar v -> copy_atom (venv_lookup venv pos v)

let eval_nocopy_atom venv pos a =
   let pos = string_pos "eval_atom" pos in
      match a with
         AtomChar c -> ValChar (ref c)
       | AtomInt i -> ValInt (ref i)
       | AtomFloat x -> ValFloat (ref x)
       | AtomNil _ -> ValPointer (ref None)
       | AtomVar v -> venv_lookup venv pos v

(*
 * General evaluator.
 *)
let rec eval_expr (tenv : tenv) (venv : venv) data_sym code =
   if !debug_eval then
      debugger venv code;
   let pos = string_pos "eval_expr" (atom_pos code) in
      match code with
         LetTypes (types, e) ->
            eval_types_expr tenv venv data_sym pos types e
       | LetFuns (funs, e) ->
            eval_funs_expr tenv venv data_sym pos funs e
       | LetAtom (v, _, a, e) ->
            eval_atom_expr tenv venv data_sym pos v a e
       | LetCopy (v, ty, a, e) ->
            eval_copy_expr tenv venv data_sym pos v ty a e
       | LetUnop (v, _, op, a, e) ->
            eval_unop_expr tenv venv data_sym pos v op a e
       | LetBinop (v, _, op, a1, a2, e) ->
            eval_binop_expr tenv venv data_sym pos v op a1 a2 e
       | LetApply (v, _, f, args, e) ->
            eval_apply_expr tenv venv data_sym pos v f args e
       | TailCall (f, args) ->
            eval_tailcall_expr tenv venv data_sym pos f args
       | LetExtCall (v, f, _, args, e) ->
            eval_apply_expr tenv venv data_sym pos v f args e
       | IntCall (_, f, args) ->
            eval_intcall_expr tenv venv data_sym pos f args
       | Return a ->
            eval_return_expr tenv venv data_sym pos a
       | LetString (v, s, e) ->
            eval_string_expr tenv venv data_sym pos v s e
       | IfThenElse (op, a1, a2, e1, e2) ->
            eval_if_expr tenv venv data_sym pos op a1 a2 e1 e2
       | SetVar (v, _, a, e) ->
            eval_set_var_expr tenv venv data_sym pos v a e
       | LetAddrOfVar (v1, _, v2, e) ->
            eval_addr_of_var_expr tenv venv data_sym pos v1 v2 e
       | SetSubscript (a1, _, a2, a3, e) ->
            eval_set_subscript_expr tenv venv data_sym pos a1 a2 a3 e
       | LetSubscript (v, _, a1, a2, e) ->
            eval_subscript_expr tenv venv data_sym pos v a1 a2 e
       | LetAddrOfSubscript (v, _, a1, a2, e) ->
            eval_addr_of_subscript_expr tenv venv data_sym pos v a1 a2 e
       | SetProject (a1, l, _, a2, e) ->
            eval_set_project_expr tenv venv data_sym pos a1 l a2 e
       | LetProject (v, _, a, l, e) ->
            eval_project_expr tenv venv data_sym pos v a l e
       | LetAddrOfProject (v, _, a, l, e) ->
            eval_addr_of_project_expr tenv venv data_sym pos v a l e
       | Memcpy (a1, _, a2, e) ->
            eval_memcpy_expr tenv venv data_sym pos a1 a2 e
       | LetAlloc (v, ty, e) ->
            eval_alloc_expr tenv venv data_sym pos v ty e
       | LetClosure (v, ty, f, args, e) ->
            eval_closure_expr tenv venv data_sym pos v f args e

(*
 * Discard the type definitions.
 *)
and eval_types_expr tenv venv data_sym pos types e =
   let pos = string_pos "eval_types_expr" pos in
   let tenv =
      List.fold_left (fun tenv (v, ty) ->
            tenv_add tenv v ty) tenv types
   in
      eval_expr tenv venv data_sym e

(*
 * Evaluate the function definitions.
 * Functions are recursive.  Make two passes,
 * collect function defs as refs, then set them in the
 * second pass.
 *)
and eval_funs_expr tenv venv data_sym pos funs e =
   (* Function declarations *)
   let pos = string_pos "eval_funs_expr" pos in
   let rec add_fun_decl venv = function
      (f, _, _, _, _) :: funs ->
         let venv = venv_add venv f (ValFun (ref None)) in
            add_fun_decl venv funs
    | [] ->
         venv
   in
   let venv = add_fun_decl venv funs in

   (* Now add the functions for real *)
   let add_fun_def (f, gflag, _, vars, body) =
      match venv_lookup venv pos f with
         ValFun p ->
            let f' _ pos args =
               let pos = string_pos "funcall" pos in
               let venv = venv_add_args venv pos vars args in
                  if gflag then
                     try eval_expr tenv venv data_sym body with
                        ReturnExn (_, v) ->
                           v
                  else
                     eval_expr tenv venv data_sym body
            in
               p := Some (List.length vars, f')
       | _ ->
            raise (Invalid_argument "Fc_ir_eval.eval_funs_expr.add_fun_def")
   in
      List.iter add_fun_def funs;
      eval_expr tenv venv data_sym e

(*
 * Atom.
 *)
and eval_atom_expr tenv venv data_sym pos v a e =
   let pos = string_pos "eval_atom_expr" pos in
   let venv = venv_add venv v (eval_atom venv pos a) in
      eval_expr tenv venv data_sym e

and eval_copy_expr tenv venv data_sym pos v ty a e =
   let pos = string_pos "eval_copy_expr" pos in
   let a =
      match a with
         Some v -> eval_atom venv pos v
       | None -> default_value tenv pos ty
   in
   let venv = venv_add venv v a in
      eval_expr tenv venv data_sym e

(*
 * Unary operation.
 *)
and eval_unop_expr tenv venv data_sym pos v op a e =
   let pos = string_pos "eval_unop_expr" pos in
   let a = eval_atom venv pos a in
   let a =
      match op, a with
         UMinusOp, ValInt { contents = i } ->
            ValInt (ref (-i))
       | UMinusOp, ValFloat { contents = x } ->
            ValFloat (ref (-.x))
       | UNotOp, ValInt { contents = i } ->
            ValInt (ref (if i = 0 then 1 else 0))
       | UNotOp, ValFloat { contents = x } ->
            ValInt (ref (if x = 0.0 then 1 else 0))
       | UNotOp, ValPointer { contents = None } ->
            ValInt (ref 1)
       | UNotOp, ValPointer { contents = Some _ } ->
            ValInt (ref 0)
       | UToChar, ValChar _ ->
            a
       | UToChar, ValInt { contents = i } ->
            ValChar (ref (Char.chr (i land 255)))
       | UToChar, ValFloat { contents = x } ->
            ValChar (ref (Char.chr ((int_of_float x) land 255)))
       | UToInt, ValChar { contents = c } ->
            ValInt (ref (Char.code c))
       | UToInt, ValInt _ ->
            a
       | UToInt, ValFloat { contents = x } ->
            ValInt (ref (int_of_float x))
       | UToFloat, ValChar { contents = c } ->
            ValFloat (ref (float_of_int (Char.code c)))
       | UToFloat, ValInt { contents = i } ->
            ValFloat (ref (float_of_int i))
       | UToFloat, ValFloat _ ->
            a
       | UToFun, ValPointer { contents = Some (a, i) } ->
            if i < 0 || i >= Array.length a then
               raise (IRException (pos, SubscriptOutOfBounds i));
            a.(i)
       | UToFunPointer, ValFun _ ->
            ValPointer (ref (Some (Array.create 1 a, 0)))
       | _ ->
            raise (EvalException (pos, venv, "eval_unop_expr: type error"))
   in
   let venv = venv_add venv v a in
      eval_expr tenv venv data_sym e

(*
 * Binary operation.
 *)
and eval_binop_expr tenv venv data_sym pos v op a1 a2 e =
   let pos = string_pos "eval_binop_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
   let a =
      match op with
         PlusOp -> eval_ptr_arith venv pos (+) (+.) a1 a2
       | MinusOp -> eval_ptr_arith venv pos (-) (-.) a1 a2
       | MulOp -> eval_arith venv pos ( * ) ( *. ) a1 a2
       | DivOp -> eval_arith venv pos (/) (/.) a1 a2
       | RemOp -> eval_int_arith venv pos (mod) a1 a2
       | BAndOp -> eval_int_arith venv pos (land) a1 a2
       | BOrOp -> eval_int_arith venv pos (lor) a1 a2
       | LslOp -> eval_int_arith venv pos (lsl) a1 a2
       | LsrOp -> eval_int_arith venv pos (lsr) a1 a2
       | AsrOp -> eval_int_arith venv pos (asr) a1 a2
       | XorOp -> eval_int_arith venv pos (lxor) a1 a2
   in
   let venv = venv_add venv v a in
      eval_expr tenv venv data_sym e

(*
 * Binary relation.
 *)
and eval_relop_flag tenv venv data_sym pos op a1 a2 =
   let pos = string_pos "eval_relop_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
      match op with
         EqOp -> eval_equal venv pos a1 a2
       | NEqOp -> not (eval_equal venv pos a1 a2)
       | LeOp -> eval_relation venv (<=) (<=) (ptr_compare (<=)) pos a1 a2
       | LtOp -> eval_relation venv (<)  (<)  (ptr_compare (<))  pos a1 a2
       | GtOp -> eval_relation venv (>)  (>)  (ptr_compare (>))  pos a1 a2
       | GeOp -> eval_relation venv (>=) (>=) (ptr_compare (>=)) pos a1 a2

(*
 * Function application.
 *)
and eval_tailcall_expr tenv venv data_sym pos f args =
   let pos = string_pos "eval_tailcall_expr" pos in
   let args = List.map (eval_atom venv pos) args in
   let f =
      match venv_lookup venv pos f with
         ValFun { contents = Some (_, f) } -> f
       | _ -> raise (EvalException (pos, venv, "eval_tailcall_expr: not a function"))
   in
      f venv pos args

and eval_apply_expr tenv venv data_sym pos v f args e =
   let pos = string_pos "eval_apply_expr" pos in
   let v' = eval_tailcall_expr tenv venv data_sym pos f args in
   let venv = venv_add venv v v' in
      eval_expr tenv venv data_sym e

and eval_intcall_expr tenv venv data_sym pos f args =
   (* Add data and exit args *)
   let args = AtomVar data_sym :: AtomVar exit_sym :: args in
      try eval_tailcall_expr tenv venv data_sym pos f args with
         ReturnExn (_, v') ->
            v'

(*
 * Function return throws an exception to the
 * enclosing function.
 *)
and eval_return_expr tenv venv data_sym pos a =
   let pos = string_pos "eval_return_expr" pos in
      raise (ReturnExn (venv, eval_atom venv pos a))

(*
 * String allocation.
 *)
and eval_string_expr tenv venv data_sym pos v s e =
   let pos = string_pos "eval_string_expr" pos in
   let s = make_string_val s in
   let venv = venv_add venv v s in
      eval_expr tenv venv data_sym e

(*
 * Conditional.
 *)
and eval_if_expr tenv venv data_sym pos op a1 a2 e1 e2 =
   let pos = string_pos "eval_if_expr" pos in
   let e =
      if eval_relop_flag tenv venv data_sym pos op a1 a2 then
         e1
      else
         e2
   in
      eval_expr tenv venv data_sym e

(*
 * Set the value of a variable.
 *)
and eval_set_var_expr tenv venv data_sym pos v a e =
   let pos = string_pos "eval_set_var_expr" pos in
   let a = eval_atom venv pos a in
   let v = venv_lookup venv pos v in
      set_value venv pos v a;
      eval_expr tenv venv data_sym e

(*
 * Address of a var.
 *)
and eval_addr_of_var_expr tenv venv data_sym pos v1 v2 e =
   let pos = string_pos "eval_addr_of_var_expr" pos in
   let v2 = venv_lookup venv pos v2 in
   let a = Array.create 1 v2 in
   let venv = venv_add venv v1 (ValPointer (ref (Some (a, 0)))) in
      eval_expr tenv venv data_sym e

(*
 * Subscripting.
 *)
and eval_subscript_expr tenv venv data_sym pos v a1 a2 e =
   let pos = string_pos "eval_subscript_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
   let a, i = coerce_pointer venv pos a1 in
   let j = coerce_int venv pos a2 in
   let j = i + j in
   let v' =
      if j < 0 || j >= Array.length a then
         raise (IRException (pos, SubscriptOutOfBounds j));
      a.(j)
   in
   let venv = venv_add venv v v' in
      eval_expr tenv venv data_sym e

and eval_set_subscript_expr tenv venv data_sym pos a1 a2 a3 e =
   let pos = string_pos "eval_set_subscript_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
   let a3 = eval_atom venv pos a3 in
   let a, i = coerce_pointer venv pos a1 in
   let j = coerce_int venv pos a2 in
   let j = j + i in
   let _ =
      if j < 0 || j >= Array.length a then
         raise (IRException (pos, SubscriptOutOfBounds j));
      set_value venv pos a.(j) a3
   in
      eval_expr tenv venv data_sym e

and eval_addr_of_subscript_expr tenv venv data_sym pos v a1 a2 e =
   let pos = string_pos "eval_addr_of_subscript_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
   let a, i = coerce_pointer venv pos a1 in
   let j = coerce_int venv pos a2 in
   let j = j + i in
   let v' = ValPointer (ref (Some (a, j))) in
   let venv = venv_add venv v v' in
      eval_expr tenv venv data_sym e

(*
 * Record projection.
 *)
and eval_project_expr tenv venv data_sym pos v a l e =
   let pos = string_pos "eval_project_expr" pos in
   let a = eval_atom venv pos a in
   let fields = coerce_struct venv pos a in
   let v' =
      try List.assoc l fields with
         Not_found ->
            raise (IRException (pos, UnboundLabel l))
   in
   let venv = venv_add venv v v' in
      eval_expr tenv venv data_sym e

and eval_set_project_expr tenv venv data_sym pos a1 l a2 e =
   let pos = string_pos "eval_set_project_expr" pos in
   let a1 = eval_nocopy_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
   let fields = coerce_struct venv pos a1 in
   let a1 =
      try List.assoc l fields with
         Not_found ->
            raise (IRException (pos, UnboundLabel l))
   in
      set_value venv pos a1 a2;
      eval_expr tenv venv data_sym e

and eval_addr_of_project_expr tenv venv data_sym pos v a l e =
   let pos = string_pos "eval_addr_of_project_expr" pos in
   let a = eval_nocopy_atom venv pos a in
   let fields = coerce_struct venv pos a in
   let a =
      try
         let v = List.assoc l fields in
            match v with
               ValPointer _ -> v
             | _ ->
                  let a = Array.create 1 v in
                     ValPointer (ref (Some (a, 0)))
      with
         Not_found ->
            raise (IRException (pos, UnboundLabel l))
   in
   let venv = venv_add venv v a in
      eval_expr tenv venv data_sym e

(*
 * IR2 memory allocation.
 *)
and eval_memcpy_expr tenv venv data_sym pos a1 a2 e =
   let pos = string_pos "eval_memcpy_expr" pos in
   let a1 = eval_atom venv pos a1 in
   let a2 = eval_atom venv pos a2 in
      memcpy venv pos a1 a2;
      eval_expr tenv venv data_sym e

and eval_alloc_expr tenv venv data_sym pos v ty e =
   let pos = string_pos "eval_alloc_expr" pos in
   let v2 = default_value tenv pos ty in
   let venv = venv_add venv v v2 in
      eval_expr tenv venv data_sym e

and eval_closure_expr tenv venv data_sym pos v f args e =
   let pos = string_pos "eval_closure_expr" pos in
   let args = List.map (eval_atom venv pos) args in
   let f =
      match venv_lookup venv pos f with
         ValFun { contents = Some (i, f) } ->
            let f' venv pos args' =
               f venv pos (args @ args')
            in
               ValFun (ref (Some (i - List.length args, f')))
       | _ ->
            raise (EvalException (pos, venv, "eval_closure_expr: not a function"))
   in
   let venv = venv_add venv v f in
      eval_expr tenv venv data_sym e

(*
 * Global function returns the environment and the value.
 *)
let eval_expr tenv venv data_sym exp =
   try tenv, venv, eval_expr tenv venv data_sym exp with
      ReturnExn (venv, v) ->
         tenv, venv, v

(*
 * Evaluate a prog.
 *)
let eval_prog tenv venv
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =

   (* Add types *)
   let tenv =
      List.fold_left (fun tenv (v, ty) ->
            tenv_add tenv v ty) tenv types
   in

   (* Body *)
   let data_sym = new_symbol_string "data_seg" in
   let exp =
      LetAlloc (data_sym, data_type,
                LetFuns (funs,
                         TailCall (init_sym, [AtomVar data_sym; AtomVar exit_sym])))
   in
   let pos = string_pos "eval_prog" (atom_pos exp) in

   (* Allocate data segment *)
   let data = default_value tenv pos data_type in
   let venv = venv_add venv data_sym data in

   (* Add strings *)
   let venv =
      List.fold_left (fun venv (v, s) ->
            let s = make_string_val s in
               venv_add venv v s) venv strings
   in
      (* Eval funs and initializer *)
      eval_expr tenv venv data_sym exp

(************************************************************************
 * CALLING MAIN
 ************************************************************************)

(*
 * Handle exceptions.
 *)
let print_exn pos venv s =
   Format.open_vbox 0;
   Fc_ir_exn.print_exn_loc (fun e ->
         Format.open_vbox tabstop;
         Format.print_string "*** Error Location:";
         Format.print_space ();
         print_expr e;
         Format.close_box ();
         Format.print_space ()) pos;
   Format.open_hvbox tabstop;
   Format.print_string "*** Error: evaluation error: ";
   Format.print_string s;
   Format.print_space ();
   Format.print_string "State is:";
   Format.print_space ();
   print_venv venv;
   Format.close_box ();
   Format.close_box ()

(*
 * Evaluate the main function.
 *)
let eval_main_raw tenv venv data_sym argv =
   let len = Array.length argv in
   let v_argv = Symbol.new_symbol_string "argv" in
   let val_argv = ValPointer (ref (Some (Array.map make_string_val argv, 0))) in
   let venv = venv_add venv v_argv val_argv in

   (* Get the name of the main function *)
   let main_sym =
      if SymbolTable.mem venv main_sym2 then
         main_sym2
      else
         main_sym1
   in

   (* Check how many args main expects *)
   let pos = TailCall (main_sym, [AtomInt len; AtomVar v_argv]) in
   let pos = string_pos "eval_main_raw" (atom_pos pos) in
   let arity =
      match venv_lookup venv pos main_sym with
         ValFun { contents = Some (arity, _) } -> arity
       | _ -> raise (EvalException (pos, venv, "main is not defined"))
   in

   (* Add the extra args *)
   let args =
      match arity with
         2 -> []
       | 3 -> [AtomVar exit_sym]
       | 4 -> [AtomVar data_sym; AtomVar exit_sym]
       | _ -> raise (EvalException (pos, venv, "main function has the wrong number of arguments"))
   in
   let exp = TailCall (main_sym, args @ [AtomInt len; AtomVar v_argv]) in
      eval_expr tenv venv data_sym exp

(*
 * Catch and print exceptions.
 *)
let eval_main tenv venv data_sym argv =
   try eval_main_raw tenv venv data_sym argv with
      EvalException (pos, venv, s) ->
         Format.set_formatter_out_channel stderr;
         Format.open_vbox 0;
         print_exn pos venv s;
         Format.close_box ();
         Format.print_newline ();
         exit 1

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
